import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from loss_funcs.umap_loss import umap,supervised_umap
import math
import time


def squared_l2_norm(x):
    flattened = x.view(x.unsqueeze(0).shape[0], -1)
    return (flattened ** 2).sum(1)


def l2_norm(x):
    return squared_l2_norm(x).sqrt()


def direkt_loss(epoch,model, model_teacher,
                x_natural,
                y,
                optimizer,transfer_loss='lbgat',use_Trades=True,
                step_size=0.003,
                epsilon=0.031,
                perturb_steps=10,
                beta=1.0,
                distance='l_inf'):
    time1 = time.time()
    # define KL-loss
    criterion_kl = nn.KLDivLoss(size_average=False)
    mse=torch.nn.MSELoss()
    ce=torch.nn.CrossEntropyLoss()
    softmax=torch.nn.Softmax(dim=1)
    model.eval()
    batch_size = len(x_natural)
    # generate adversarial example
    x_adv = x_natural.detach() + 0.001 * torch.randn(x_natural.shape).cuda().detach()
    if distance == 'l_inf':
        for _ in range(perturb_steps):
            x_adv.requires_grad_()
            with torch.enable_grad():
                if use_Trades:
                    loss_kl = criterion_kl(F.log_softmax(model(x_adv)[1], dim=1),
                                           F.softmax(model(x_natural)[1], dim=1))
                else:
                    loss_kl = ce(model(x_adv)[1],y)
            grad = torch.autograd.grad(loss_kl, [x_adv])[0]
            x_adv = x_adv.detach() + step_size * torch.sign(grad.detach())
            x_adv = torch.min(torch.max(x_adv, x_natural - epsilon), x_natural + epsilon)
            x_adv = torch.clamp(x_adv, 0.0, 1.0)
    elif distance == 'l_2':
        for _ in range(perturb_steps):
            x_adv.requires_grad_()
            with torch.enable_grad():
                if use_Trades:
                    loss_kl = criterion_kl(F.log_softmax(model(x_adv)[1], dim=1),
                                       F.softmax(model(x_natural)[1], dim=1))
                else:
                    loss_kl = ce(model(x_adv)[1],y)
            grad = torch.autograd.grad(loss_kl, [x_adv])[0]
            for idx_batch in range(batch_size):
                grad_idx = grad[idx_batch]
                grad_idx_norm = l2_norm(grad_idx)
                grad_idx /= (grad_idx_norm + 1e-8)
                x_adv[idx_batch] = x_adv[idx_batch].detach() + step_size * grad_idx
                eta_x_adv = x_adv[idx_batch] - x_natural[idx_batch]
                norm_eta = l2_norm(eta_x_adv)
                if norm_eta > epsilon:
                    eta_x_adv = eta_x_adv * epsilon / l2_norm(eta_x_adv)
                x_adv[idx_batch] = x_natural[idx_batch] + eta_x_adv
            x_adv = torch.clamp(x_adv, 0.0, 1.0)
    else:
        x_adv = torch.clamp(x_adv, 0.0, 1.0)
    model.train()

    x_adv = Variable(torch.clamp(x_adv, 0.0, 1.0), requires_grad=False)
    # zero gradient
    optimizer.zero_grad()
    lambd = 2 / (1 + math.exp(-10 * (epoch) / 100)) - 1

    # calculate robust loss
    feature_adv, out_adv, = model(x_adv)
    feature_natural, out_natural = model(x_natural)
    feature, out = model_teacher(x_natural)
    out_teacher = out.detach()
    # feature = feature.detach()
    if transfer_loss == 'lbgat':
        lbgat = mse(out_adv,out)
        transfer_loss = lbgat
        # lbgat =  ce(out_natural,y)
        loss_mse = transfer_loss +ce(out,y)
        # loss_mse=transfer_loss+ ce(out, y)
    elif transfer_loss == 'umap':
        transfer_loss = 0.5*umap(feature_adv, feature)
        # a = transfer_loss
        lbgat = 1*mse(out_adv,out)
        loss_mse = transfer_loss+ce(out, y)+lbgat#+0*ce(out_natural,y) # +0.1*ce(out_natural,y) #+lbgat
        # loss_mse = 20 * lambd * supervised_umap(feature_adv, feature,y) + ce(out, y)
    elif transfer_loss =='supervised_umap' :
        sample_weight = torch.nn.CrossEntropyLoss(reduction='none')(out,y)
        transfer_loss = 20*lambd*supervised_umap(feature_adv, feature, y,sample_weight)
        lbgat = 1.0*mse(out_adv,out)
        loss_mse = transfer_loss + ce(out, y)+0.1*ce(out_natural,y)
    elif transfer_loss == 'pkt':
        transfer_loss = 5*cosine_similarity_loss(feature_adv, feature)
        lbgat = 1.0 * mse(out_adv, out)
        loss_mse = transfer_loss + ce(out, y) + 0.1 * ce(out_natural, y)
    else:
        loss_mse =  ce(out,y)
    loss_kl = (1.0 / out.size(0)) * criterion_kl(F.log_softmax(out_adv, dim=1),F.softmax(out_natural, dim=1))

    if use_Trades:
        robust_loss = beta * loss_kl #+ce(out_adv,y)

        loss = loss_mse + robust_loss
        # if transfer_loss == 'umap' :
        #     # loss = loss_mse + beta * loss_kl +ce(out_natural,y)
        #     loss = loss_mse + beta * loss_kl

    else:
        robust_loss = ce(out_adv,y)
        loss = loss_mse + robust_loss
    # return transfer_loss.item()*y.shape[0], loss,lbgat.item()*y.shape[0],robust_loss.item()*y.shape[0]
    return transfer_loss.item() * y.shape[0], loss, lbgat.item() * y.shape[0], robust_loss.item() * y.shape[0]
